#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>
#include <limits.h>

#include "common.h"
#include "macros.h"
#include "xmalloc.h"
#include "mlex.h"
#include "misc.h"
#include "preproc.h"
#include "keyword.h"
#include "syntax.h"
#include "symbol.h"
#include "as_tree.h"
#include "tree2text.h"
#include "list.h"
#include "quad.h"
#include "tree2inter.h"
#include "machine_internal.h"
#include "machine_generic.h"
#include "machine_vmlang.h"
#include "verify.h"
#include "peephole.h"

extern char *__progname;
static FILE *fs;

static void code_print_func(const char *fmt, ...)
{
	va_list ap;

	return_if_fail(fs != NULL);

	va_start(ap, fmt);
	vfprintf(fs, fmt, ap);
	va_end(ap);
}

static void print_usage()
{
	printf("Usage: %s [options] file\n"
	" -h		print this help\n"
	" -o <file>	place the output into <file>\n"
	" -i		use `internal' implementation\n"
	" -I		folder with include files\n"
	" -s		print the listing on the screen\n", __progname);
}

#define	OPTION_INTERNAL_ARCH	0x01
#define OPTION_USE_STDOUT	0x02
#define OPTION_OUTPUT_NAME	0x04
#define OPTION_INCLUDE_FOLDER	0x08

int main(int argc, char **argv)
{
	struct ast_node_root *root;
	char *file_name, *template, *mem, *tmp;
	char *out_name = "file.S";
	char inc_folder[PATH_MAX];
	void (*print_out_ptr)(const char *fmt, ...);
	size_t len;
	ret_t ret;
	int fd, opt, options;

	struct quad **quads;
	unsigned int quads_size, i;

	if (argc < 2) {
		print_usage();
		exit(1);
	}

	inc_folder[0] = 0;
	options = 0;
	tmp = NULL;

	while ((opt = getopt(argc, argv, "hiso:I:")) != -1) {
		switch (opt) {
		case 'h':
			print_usage();
			exit(0);
			break;
		case 'o':
			options |= OPTION_OUTPUT_NAME;
			out_name = optarg;
			break;
		case 'i':
			options |= OPTION_INTERNAL_ARCH;
			break;
		case 's':
			options |= OPTION_USE_STDOUT;
			break;
		case 'I':
			options |= OPTION_INCLUDE_FOLDER;
			tmp = canonize_path(optarg);
			break;
		default:
			print_usage();
			exit(1);
		}
	}

	argc -= optind;
	argv += optind;

	if (tmp != NULL) {
		xstrncpy(inc_folder, tmp, sizeof(inc_folder)-1);
		len = strlen(inc_folder);
		if (inc_folder[len-1] != '/') {
			inc_folder[len] = '/';
			inc_folder[len+1] = 0;
		}
		xfree(tmp);
	}

	if ((options & OPTION_OUTPUT_NAME) && (options & OPTION_USE_STDOUT)) {
		fprintf(stderr, "can't use -o and -s options together\n");
		exit(-1);
	}

	if (argc != 1) {
		fprintf(stderr, "wait for one input file %d\n", argc);
		exit(-1);
	}

	file_name = argv[0];

	asprintf(&template, "%s%s", file_name, "XXXXXX");

	fd = mkstemp(template);
	if (fd == -1) {
		xfree(template);
		fprintf(stderr, "can't create temporary file\n");
		exit(-1);
	}

	keyword_table_create();

	ret = preproc_run(file_name, fd, inc_folder);
	if (ret != 0) {
		fprintf(stderr, "preprocessor return %d errors\n", ret);
		goto free2;
	}

	ret = file_length(fd, &len);
	if (ret != ret_ok) {
		fprintf(stderr, "can't get file length `%s'\n", file_name);
		goto free2;
	}

	ret = mmap_file(fd, (void **)&mem, len);
	if (ret != ret_ok) {
		fprintf(stderr, "mmap file failed `%s'\n", file_name);
		goto free2;
	}

	ret = close_file(&fd);
	if (ret != ret_ok) {
		fprintf(stderr, "can't close file %s\n", file_name);
		goto free2;
	}

	mlex_file_setup(mem, mem + len);
	symbol_table_create_global_table();

	ret = programme(&root);
	if (ret != 0) {
		fprintf(stderr, "\nsyntaxer return %d errors\n", ret);
		goto free1;
	}

	ret = verify_nerrors();
	if (ret != 0) {
		fprintf(stderr, "verifier return %d errors\n", ret);
		goto free1;
	}

	//tree2text(root, NULL);
	print_out_ptr = NULL;

	if (options & OPTION_OUTPUT_NAME) {
		fs = fopen(out_name, "w+");
		if (fs == NULL) {
			fprintf(stderr, "can't create output file %s\n", out_name);
			goto free1;
		}
		print_out_ptr = code_print_func;
	}

	quads = tree2inter(root, &quads_size);
	peephole_optimization(quads, &quads_size);

	if (options & OPTION_INTERNAL_ARCH)
		machine_internal_set_vfuncs();
	else
		machine_vmlang_set_vfuncs();

	machine_gen_code(quads, quads_size, print_out_ptr);

	for (i = 0; i < quads_size; i++) 
		quad_unref(quads[i]);

	xfree(quads);

	if (fs != NULL)
		fclose(fs);

	ast_node_unref(AST_NODE(root));

	mlex_file_finalize();
	keyword_table_destroy();
	remove(template);
	xfree(template);

	return 0;

free1:
	ast_node_unref(AST_NODE(root));

free2:
	mlex_file_finalize();
	keyword_table_destroy();
	remove(template);
	xfree(template);

	return 1;
}

